# -*- coding: utf-8 -*-
import numpy as np
import networkx as nx
import geopandas as gpd
from shapely.geometry import Point, MultiPoint, LineString, MultiLineString
from scipy.spatial import cKDTree
from collections import defaultdict
from typing import Union, Iterable
import geonetworkx.settings as settings


PointCoordinatesLike = Iterable[float]
PointsCoordinatesLike = Union[Iterable[PointCoordinatesLike], MultiPoint]


class Extremity:
    """Represents an extremity of a line. It's useful to parse and deal with lines given as input."""
    nb_extremity = 0

    def __init__(self, shape_id, position, coords):
        self.id = Extremity.nb_extremity
        Extremity.nb_extremity += 1
        self.shape_id = shape_id
        self.position = position
        self.coords = coords
        self.related_extremity = None
        self.close_extremities = []
        self.closest_station = None
        self.is_merged = False
        self.is_added_as_shape = False
        self.matching_items = dict()


def merge_two_shape(e1: Extremity, e2: Extremity, line1: LineString, line2: LineString) -> LineString:
    """Merge two lines (line1 and line2) with the given extremities (e1 and e2)."""
    if e1.position == -1 and e2.position == 0:
        new_line = LineString(list(line1.coords) + list(line2.coords))
    elif e1.position == 0 and e2.position == 0:
        new_line = LineString(list(reversed(line1.coords)) + list(line2.coords))
    elif e1.position == 0 and e2.position == -1:
        new_line = LineString(list(reversed(line1.coords)) + list(reversed(line2.coords)))
    elif e1.position == -1 and e2.position == -1:
        new_line = LineString(list(line1.coords) + list(reversed(line2.coords)))
    else:
        assert False, 'Invalid position attribute'
    return new_line


def merge_two_lines_with_closest_extremities(first_line: LineString, second_line: LineString) -> LineString:
    """Merge two lines with their closest extremities. Euclidian distance is used here."""
    combinations = [(0, 0), (-1, 0), (0, -1), (-1, -1)]
    extremities = np.array([[first_line.coords[i1], second_line.coords[i2]] for (i1, i2) in combinations])
    distances = np.linalg.norm(extremities[:, 0, :] - extremities[:, 1, :], axis=1)
    merge_couple_index = int(np.argmin(distances))
    merge_couple = combinations[merge_couple_index]
    if merge_couple == (-1, 0):
        merged_line = LineString(list(first_line.coords) + list(second_line.coords))
    elif merge_couple == (0, 0):
        merged_line = LineString(list(reversed(first_line.coords)) + list(second_line.coords))
    elif merge_couple == (0, -1):
        merged_line = LineString(list(reversed(first_line.coords)) + list(reversed(second_line.coords)))
    else:  # merge_couple == (-1, -1)
        merged_line = LineString(list(first_line.coords) + list(reversed(second_line.coords)))
    return merged_line


def get_shape_extremities(shape: LineString, shape_id: int):
    """Return the extremities of a shape in the network_shapes_gdf."""
    assert isinstance(shape, LineString), "Shape must be a LineString"
    first_vertex = shape.coords[0]
    last_vertex = shape.coords[-1]
    e1 = Extremity(shape_id, 0, first_vertex)
    e2 = Extremity(shape_id, -1, last_vertex)
    e1.related_extremity = e2
    e2.related_extremity = e1
    return e1, e2


def convert_multilinestring_to_linestring(gdf: gpd.GeoDataFrame) -> int:
    """Convert all geometry attribute being a 'MultiLineString' to a 'LineString'. The created line is a merge of all
     sublines.

    Parameters
    ----------
    gdf : gpd.GeoDataFrame
        A GeoDataFrame with a 'geometry' column to modify

    Returns
    -------
    int
        The number of converted 'MultiLineString'

    Raises
    ------
    RuntimeError
        If an input shape is not a LineString or a MultiLineString

    """
    nb_converted_multilinestring = 0
    for s in gdf.index:
        if isinstance(gdf.at[s, 'geometry'], LineString):
            continue
        elif isinstance(gdf.at[s, 'geometry'], MultiLineString):
            all_lines = gdf.at[s, 'geometry'].geoms
            new_line = all_lines[0]
            for i in range(1, len(all_lines)):
                new_line = merge_two_lines_with_closest_extremities(new_line, all_lines[i])
            gdf.at[s, 'geometry'] = new_line
            nb_converted_multilinestring += 1
        else:
            raise RuntimeError("Unknown shape type for shape input at index : ", s)
    return nb_converted_multilinestring


def discretize_line(line: LineString) -> list:
    """Takes a shapely LineString and discretize it into a list of shapely Points. Each point is at most at the
    discretization tolerance distance of the following point.

    Parameters
    ----------
    line : LineString
        Line to discretize

    Returns
    -------
    list
        An ordered list of shapely Point

    See Also
    --------
    discretize_lines

    """
    points_list = []
    current_dist = settings.DISCRETIZATION_TOLERANCE
    line_length = line.length
    points_list.append(Point(list(line.coords)[0]))
    while current_dist < line_length:
        points_list.append(line.interpolate(current_dist))
        current_dist += settings.DISCRETIZATION_TOLERANCE
    points_list.append(Point(list(line.coords)[-1]))
    return points_list


def discretize_lines(lines: Iterable[LineString]):
    """Discretize some line into points.

    Parameters
    ----------
    lines: Iterable[LineString] :
        Lines to discretize

    Returns
    -------
    MultiPoint and defaultdict
        Return all the discretized points as a shapely MultiPoint and a dictionary to map the discretized points for
        each line.

    See Also
    --------
    discretize_line

    """
    points_line_association = defaultdict(list)
    all_points = []
    nb_points = 0
    for line_index, line in enumerate(lines):
        discretized_points = discretize_line(line)
        nb_discretized_points = len(discretized_points)
        all_points.extend(discretized_points)
        points_line_association[line_index] = list(range(nb_points, nb_points + nb_discretized_points))
        nb_points += nb_discretized_points
    all_points_as_mp = MultiPoint(all_points)
    return all_points_as_mp, points_line_association


def get_closest_point_from_points(points_from: PointsCoordinatesLike, points_to: list = None, kd_tree: cKDTree = None):
    """Compute the closest point among the ``points_from`` list for each point in the ``points_to`` list.

    Parameters
    ----------
    points_from : PointsCoordinatesLike
        Iterable of points coordinates
    points_to : list or None
        Iterable of points coordinates (Default value = None)
    kd_tree : cKDTree
        a constructed kd tree representing ``points_from`` (Default value = None)

    Returns
    -------
    array of floats
        distances
    ndarray of ints
        indexes

    """
    if points_to is None and kd_tree is None:
        raise ValueError("Must provide at least argument 'points_to' or 'kd_tree'")
    if kd_tree is None:
        kd_tree = cKDTree(points_to)
    return kd_tree.query(points_from)


def get_closest_point_from_line(line_from: LineString, points_to: list = None, kd_tree: cKDTree = None):
    """Return the closest point from a given line and its distance.

    Parameters
    ----------
    line_from : LineString
        A shapely LineString (Default value = None)
    points_to : list
        A list of points among which the closest to the line has to be found (optional is ``kdtree`` is
        given)
    kd_tree : cKDTree
        A kd-tree representing the points among which the closest to the line has to be found (optional if
        ``points_to`` is given) (Default value = None)

    Returns
    -------
    float
        closest distance
    int
        index of the closest point

    """
    if points_to is None and kd_tree is None:
        raise ValueError("Must provide at least argument 'points_to' or 'kd_tree'")
    if kd_tree is None:
        kd_tree = cKDTree(points_to)
    discretized_points = discretize_line(line_from)
    distances, closest_points_indexes = kd_tree.query(MultiPoint(discretized_points))
    smallest_distance_index = np.argmin(distances)
    return distances[smallest_distance_index], closest_points_indexes[smallest_distance_index]


def get_closest_point_from_multi_shape(multi_shape, points_to=None, kd_tree=None):
    """Computes the closest point to the multi shape (i.e. the point that has the smallest projection distance on the
    entire multi shape object.

    Parameters
    ----------
    multi_shape :  MultiPoint or MultiLineString
        The multi shape object can be any shapely object among: MultiPoint, MultiLineString
    points_to : list
        A list of points among which to find the closest to the multi shape (Default value = None)
    kd_tree : cKDTree
        A kdtree representing the points among which the closest to the multishape has to be found (optional
        if 'points_to' is given) (Default value = None)

    Returns
    -------
    float
        distance
    int
        index of the closest point

    """
    if isinstance(multi_shape, MultiPoint):
        distances_and_points_indexes = get_closest_point_from_points(multi_shape, points_to, kd_tree)
    elif isinstance(multi_shape, MultiLineString):
        distances_and_points_indexes = []
        for line in multi_shape:
            distances_and_points_indexes.append(get_closest_point_from_line(line, points_to, kd_tree))
    else:
        raise TypeError("Method not implemented for shape of type '%s', expected MultiPoint"
                        "or MultiLineString" % str(type(multi_shape)))
    smallest_distance_index = np.argmin(d_ix[0] for d_ix in distances_and_points_indexes)
    return distances_and_points_indexes[smallest_distance_index]


def get_closest_point_from_shape(shape: Union[Point, LineString, MultiPoint, MultiLineString],
                                 points_to: Union[MultiPoint, np.ndarray, list] = None,
                                 kd_tree: cKDTree = None):
    """Compute the closest point to the given shape.

    Parameters
    ----------
    shape : Point, MultiPoint, LineString or MultiLineString
        Any shapely shape
    points_to : list
        A list of points among which to find the closest to the multi shape (Default value = None)
    kd_tree : cKDTree
        A kdtree representing the points among which the closest to the shape has to be found (optional if
        'points_to' is given) (Default value = None)

    Returns
    -------
    float
        distance
    int
        index of the closest point

    """
    if isinstance(shape, Point):
        result = get_closest_point_from_points(MultiPoint([shape]), points_to, kd_tree=kd_tree)
        return result[0][0], result[1][0]
    elif isinstance(shape, LineString):
        return get_closest_point_from_line(shape, points_to, kd_tree=kd_tree)
    elif isinstance(shape, (MultiPoint, MultiLineString)):
        return get_closest_point_from_multi_shape(shape, points_to, kd_tree=kd_tree)
    else:
        raise TypeError("Method not implemented for shape of type '%s', expected Point, "
                        "LineString or MultiLineString" % str(type(shape)))


def get_closest_point_from_shapes(shapes_from, points_to):
    """Compute the closest point for each given shape.

    Parameters
    ----------
    shapes_from :
        An iterable of shapes (Point, MultiPoint, LineString, MultiLineString)
    points_to : list
        A list of points among which to find the closest to the multi shape

    Returns
    -------
    float
        distance
    int
        index of the closest point

    """
    results = []
    kd_tree = cKDTree(points_to)
    for shape in shapes_from:
        distance, closest_points_index = get_closest_point_from_shape(shape, kd_tree=kd_tree)
        results.append((distance, closest_points_index))
    return results


def get_closest_line_from_point(point_from: PointCoordinatesLike,
                                lines_to=None,
                                kd_tree=None,
                                points_line_association=None):
    """Find the closest line from a given point.

    Parameters
    ----------
    point_from : PointCoordinatesLike
        Point coordinate to find the closest line.
    lines_to : list
        Group of lines among which the closest has to be found (optional if ``kdtree`` and
        ``points_line_association`` are given). (Default value = None)
    kd_tree : cKDTree
        An optional pre-computed kd_tree of discretized lines. (Default value = None)
    points_line_association : dict
        An optional pre-computed dictionary matching lines and discretized points. (Default value = None)

    Returns
    -------
    float
        distance
    int
        index of the closest line

    """
    if lines_to is None and kd_tree is None:
        raise ValueError("Must provide at least argument 'points_to' or 'kd_tree'")
    elif kd_tree is None and points_line_association is None:
        raise ValueError("If a kd-tree is given, a point line association dictionary must provided")
    if kd_tree is None:
        points_to, points_line_association = discretize_lines(lines_to)
        kd_tree = cKDTree(points_to)
    distance, closest_point_index = get_closest_point_from_points([point_from], kd_tree=kd_tree)
    for line_index in points_line_association:
        if closest_point_index in points_line_association[line_index]:
            return distance, line_index
    assert False


def get_closest_line_from_points(points_from, lines_to):
    """Find the closest line for each given points.

    Parameters
    ----------
    points_from : list
        Points coordinates.
    lines_to : list
        Group of lines among which the closest has to be found.

    Returns
    -------
    list
        A list of closest lines indexes.

    """
    points_to, points_line_association = discretize_lines(lines_to)
    kd_tree = cKDTree(points_to)
    lines_indexes = []
    for point in points_from:
        result = get_closest_line_from_point(point, kd_tree=kd_tree, points_line_association=points_line_association)
        lines_indexes.append(result[1])
    return lines_indexes


def get_polygons_neighborhood(polygons):
    """Returns for each polygon a set of intersecting polygons."""
    nb_polygons = len(polygons)
    neighborhoods = [set() for c in range(nb_polygons)]
    for ix1, p1 in enumerate(polygons):
        for ix2 in range(ix1 + 1, nb_polygons):
            p2 = polygons[ix2]
            if p1.intersects(p2):
                neighborhoods[ix1].add(ix2)
                neighborhoods[ix2].add(ix1)
    return neighborhoods


def split_line(line, distance):
    """Cuts a line in two at a distance from its starting point."""
    coords = list(line.coords)
    if distance <= 0.0:
        return [LineString([coords[0]]*2), LineString(line)]
    if distance >= line.length:
        return [LineString(line), LineString([coords[-1]]*2)]
    for i, p in enumerate(coords):
        pd = line.project(Point(p))
        if pd == distance:
            return [LineString(coords[:i + 1]), LineString(coords[i:])]
        if pd > distance:
            cp = line.interpolate(distance)
            return [LineString(coords[:i] + [(cp.x, cp.y)]), LineString([(cp.x, cp.y)] + coords[i:])]


def coordinates_almost_equal(c1: Iterable, c2: Iterable, tolerance=1e-8) -> bool:
    """Returns true if the two given list of coordinates equals within a given tolerance.

    Parameters
    ----------
    c1: Iterable
        First point coordinates
        
    c2: Iterable
        Second point coordinates
        
    tolerance : float
         Tolerance comparison (Default value = 1e-8)

    Returns
    -------
    bool
        True if the coordinates almost equal, false otherwise.

    """
    for i, j in zip(c1, c2):
        if abs(i - j) > tolerance:
            return False
    return True


def almost_equally_located(p1: Point, p2: Point, tolerance=1e-8) -> bool:
    """Test if two point are loacated at the same place within a tolerance.

    Parameters
    ----------
    p1: Point
        First point to compare
    p2: Point
        Second point to compare
    tolerance :
         Comparison tolerance (Default value = 1e-8)

    Returns
    -------
    bool
        True if the two points have the same coordinates.
    """
    return coordinates_almost_equal([p1.x, p1.y], [p2.x, p2.y], tolerance)


def insert_point_in_line(line: LineString, point_coords: list, position: int) -> LineString:
    """Insert a new point in a line given its coordinates."""
    new_line_coordinates = line.coords[0:position]
    new_line_coordinates.append((point_coords[0], point_coords[1]))
    new_line_coordinates.extend(line.coords[position:])
    return LineString(new_line_coordinates)



# -*- coding: utf-8 -*-
from shapely.geometry import LineString
import numpy as np
from nose.plugins.attrib import attr
import unittest
from geonetworkx.geometry_operations import discretize_lines
from geonetworkx.testing.utils import assert_almost_intersect
import geonetworkx.testing.utils as gnx_tu
import geonetworkx.settings as settings


gnx_tu.SEED = 70595
np.random.seed(gnx_tu.SEED)


@attr('geometry_operations')
class TestGeometryOperations(unittest.TestCase):

    def test_discretize_lines(self):
        # Test that a discretization of lines is well set for each line
        settings.DISCRETIZATION_TOLERANCE = 1e-1
        nb_lines = 20
        lines = []
        for l in range(nb_lines):
            line_nb_points = np.random.randint(5, 30)
            points = np.random.rand(line_nb_points, 2)
            lines.append(LineString(points))
        discretized_points, points_line_association = discretize_lines(lines)
        for l in points_line_association:
            line = lines[l]
            for p in points_line_association[l]:
                point = discretized_points[p]
                assert_almost_intersect(point, line, "A discretized point does not intersects its given line")





# -*- coding: utf-8 -*-
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fr:
    install_requires = fr.read().splitlines()

setuptools.setup(
    name="geonetworkx",
    version="0.4",
    author="Hugo Chareyre - Artelys",
    author_email="hugo.chareyre@artelys.com",
    description="Python tools for geographic graphs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ])



# -*- coding: utf-8 -*-
import networkx as nx
import geonetworkx as gnx
import unittest
from nose.tools import assert_true, assert_equal, assert_less_equal
from nose.plugins.attrib import attr
import geonetworkx.testing.utils as gnx_tu
from geonetworkx.utils.voronoi_utils import *
from geonetworkx.generators import extended_ego_graph
from geonetworkx.tests import datasets
import os

gnx_tu.SEED = 70595
NB_POINTS = 50
data_directory = os.path.join(os.path.dirname(os.path.realpath(__file__)), "datasets")


@attr('tools')
class TestUtils(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_voronoi_edges(self):
        gmg = datasets.get_grenoble_streets_200()
        gnx.fill_edges_missing_geometry_attributes(gmg)
        edge_as_lines = gmg.get_edges_as_line_series()
        lines = list(edge_as_lines)
        tolerance = 1e-7
        res = compute_voronoi_cells_from_lines(lines, tolerance)
        for e, line in edge_as_lines.items():
            cell_found = False
            for p in res:
                if p.buffer(10 * tolerance).contains(line):
                    cell_found = True
                    break
            if not cell_found:
                assert_true(False, "A edge geometry '%s' is not in any voronoi cells" % str(e))

    def test_extented_ego_graph(self):
        gmg = datasets.get_grenoble_streets_200()
        gnx.fill_edges_missing_geometry_attributes(gmg)
        gnx.fill_length_attribute(gmg, "length", only_missing=True)
        source = 312173744
        limit = 100  # meters
        ego_graph = extended_ego_graph(gmg, source, radius=limit, center=True, undirected=False, distance="length")
        ccs = list(nx.connected_components(ego_graph))
        assert_equal(len(ccs), 1, "The ego graph has several connected components")
        path_lengths = nx.single_source_dijkstra_path_length(ego_graph, source, weight="length")
        for n, length in path_lengths.items():
            assert_less_equal(length, limit, "A path in the ego graph is too long: %f > %f" % (length, limit))



# -*- coding: utf-8 -*-
import os
import shutil
import geopandas as gpd
import networkx as nx
import geonetworkx as gnx
import numpy as np
from shapely.geometry import Point
from nose.tools import assert_in, assert_less
import unittest
from nose.plugins.attrib import attr
import geonetworkx.testing.utils as gnx_tu
from geonetworkx.tests import datasets

gnx_tu.SEED = 70595
data_directory = os.path.join(os.path.dirname(os.path.realpath(__file__)), "datasets")


@attr('tools')
class TestTools(unittest.TestCase):

    def setUp(self):
        file_dir = os.path.dirname(__file__)
        self.results_dir = os.path.join(file_dir, "datasets/results/")
        if not os.path.exists(self.results_dir):
            os.mkdir(self.results_dir)

    def tearDown(self):
        shutil.rmtree(self.results_dir)

    def test_spatial_points_merge(self):
        # test a spatial merge
        graph = datasets.get_copenhagen_street_net()
        gnx.utils.fill_edges_missing_geometry_attributes(graph)
        points_gdf = datasets.get_copenhagen_buildings()
        gnx.tools.spatial_points_merge(graph, points_gdf, inplace=True)
        for p in points_gdf.index:
            assert_in(p, graph.nodes())

    def test_spatial_graph_merge(self):
        base_graph = datasets.get_copenhagen_street_net().to_undirected()
        gnx.utils.fill_edges_missing_geometry_attributes(base_graph)
        base_graph.graph["name"] = "streets"

        other_graph = datasets.get_copenhagen_ferry_net().to_undirected()
        gnx.order_well_lines(other_graph)
        gnx.join_lines_extremity_to_nodes_coordinates(other_graph)
        other_graph.graph["name"] = "ferry"
        merged_graph = gnx.tools.spatial_graph_merge(base_graph, other_graph, inplace=False)
        merged_graph.graph["name"] = "merged_graph"
        for n in other_graph.nodes():
            assert_in(n, merged_graph.nodes())
        for n in base_graph.nodes():
            assert_in(n, merged_graph.nodes())
        for e in other_graph.edges:
            assert_in(e, merged_graph.edges)

    def test_spatial_graph_merge_with_non_distinct_nodes(self):
        # base graph definition
        nb_nodes = 20
        base_graph = gnx_tu.get_random_geograph(nb_nodes)
        gnx.fill_edges_missing_geometry_attributes(base_graph)

        # other graph definition
        nb_nodes_2 = 20
        edge_creation_prob = 0.1
        other_graph = gnx_tu.get_random_geograph(nb_nodes_2)
        nx.relabel_nodes(other_graph, {n: n + nb_nodes for n in other_graph.nodes}, False)
        distinct_nodes = list(other_graph.nodes())
        nx.set_node_attributes(other_graph, 2, "origin")
        common_nodes = list(range(nb_nodes // 4))
        common_part = nx.subgraph(base_graph, common_nodes)
        other_graph = nx.compose(other_graph, common_part)
        for n1 in common_nodes:
            for n2 in distinct_nodes:
                if np.random.rand() <= edge_creation_prob:
                    other_graph.add_edge(n1, n2)
        gnx.fill_edges_missing_geometry_attributes(other_graph)
        nx.set_node_attributes(other_graph, {n: 3 for n in common_nodes}, "origin")
        nx.set_node_attributes(base_graph, {n: 3 for n in common_nodes}, "origin")
        # nx.draw_networkx(base_graph, pos=nodes_coords, node_color ="red")
        # nx.draw_networkx(other_graph, pos=other_graph.get_nodes_coordinates(),
        #                  node_color=[2 if n in base_graph.nodes() else 3 for n in g2.nodes()])

        merged_graph = gnx.spatial_graph_merge(base_graph, other_graph, inplace=False)
        # origins = nx.get_node_attributes(merged_graph, "origin")
        # nc = [origins[n] if n in origins else 4 for n in merged_graph.nodes()]
        # nx.draw_networkx(merged_graph, pos=merged_graph.get_nodes_coordinates(), node_color=nc)

    def test_spatial_point_merge_with_filters(self):
        nb_nodes = 30
        graph = gnx_tu.get_random_geograph(nb_nodes)
        gnx.fill_edges_missing_geometry_attributes(graph)
        nb_points = 200
        points = np.random.rand(nb_points, 2)
        points_gdf = gpd.GeoDataFrame(columns=["geometry"])
        points_gdf["geometry"] = [Point([x, y]) for x, y in points]
        points_gdf.index = range(nb_nodes, nb_nodes + nb_points)
        # node filter
        node_filter = lambda n: n > 5
        merged_graph = gnx.spatial_points_merge(graph, points_gdf, node_filter=node_filter, inplace=False)
        for n in graph.nodes:
            assert_in(n, merged_graph.nodes)
        for n in points_gdf.index:
            assert_in(n, merged_graph.nodes)
        for n in graph.nodes:
            if not node_filter(n):
                edges = [e for e in graph.edges if n in e]
                for e in edges:
                    assert_in(e, merged_graph.edges)
        # edge filter
        edge_filter = lambda u, v: u not in range(5, 10) and v not in range(12, 20)
        merged_graph = gnx.spatial_points_merge(graph, points_gdf, edge_filter=edge_filter, inplace=False)
        for n in graph.nodes:
            assert_in(n, merged_graph.nodes)
        for n in points_gdf.index:
            assert_in(n, merged_graph.nodes)
        for e in graph.edges:
            if not edge_filter(e[0], e[1]):
                assert_in(e, merged_graph.edges)
        # node filter and edge filter
        merged_graph = gnx.spatial_points_merge(graph, points_gdf, edge_filter=edge_filter, node_filter=node_filter,
                                                inplace=False)
        for n in graph.nodes:
            assert_in(n, merged_graph.nodes)
        for n in points_gdf.index:
            assert_in(n, merged_graph.nodes)
        for e in graph.edges:
            if not edge_filter(e[0], e[1]):
                assert_in(e, merged_graph.edges)
        for n in graph.nodes:
            if not node_filter(n):
                edges = [e for e in graph.edges if n in e]
                for e in edges:
                    assert_in(e, merged_graph.edges)

    def test_isochrone(self):
        # Read data
        gmg = datasets.get_grenoble_streets_500()
        gnx.fill_edges_missing_geometry_attributes(gmg)
        gnx.fill_length_attribute(gmg, "length", True)

        # Compute Isochrone polygon
        source = 312173744
        limit = 400  # meters
        iso_polygon = gnx.isochrone_polygon(gmg, source, limit, "length")

        # Try to compute the distance for a set of point that are in the polygon bounding box
        nb_test_points = 20
        np.random.seed(gnx_tu.SEED)
        bb = iso_polygon.bounds
        points_coords = np.random.rand(nb_test_points, 2) * np.array([bb[2] - bb[0], bb[3] - bb[1]]) + \
                        np.array([bb[0], bb[1]])
        points = gpd.GeoDataFrame([Point(p) for p in points_coords], columns=["geometry"], crs=gmg.crs)
        merged_graph = gnx.spatial_points_merge(gmg, points_gdf=points, inplace=False)
        gnx.fill_length_attribute(merged_graph, "length", True)
        tol = 0.05 * limit
        for p in points.index:
            shortest_distance = nx.shortest_path_length(merged_graph, source, p, weight="length")
            if points.at[p, "geometry"].intersects(iso_polygon):
                assert_less(shortest_distance - tol, limit, "The shortest path between a point intersecting the iso-polygon"
                                                            " must be shorter than the isochrone limit.")
            else:
                assert_less(limit, shortest_distance + tol, "The shortest path between a point not intersecting the "
                                                            "iso-polygon must be longer than the isochrone limit.")
